from django.contrib.auth.models import User
from django.db import models
from resources.models import Resource


class Annotation(models.Model):
    resource = models.ForeignKey(Resource, on_delete=models.CASCADE)
    owner = models.ForeignKey(User, on_delete=models.CASCADE)
    data = models.JSONField("JSON data")
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)



from django.core.management.base import BaseCommand
from inspect import signature
import re
from rpc.views import _get_rpc_methods

all_methods = _get_rpc_methods()
under_pat = re.compile(r"_([a-z])")


def underscore_to_camel(name):
    return under_pat.sub(lambda x: x.group(1).upper(), name)


PARAM_TPL = "     * @param {} {}"
RETURN_TPL = "     * @returns {}"
FN_TPL = """
    /**
{}
     */
    async {}({}) {{
        return this._rpc("{}", [{}]);
    }}
"""


types_trans = {
    "<class 'str'>": "{String}",
    "<class 'int'>": "{BigInteger}",
    "<class 'bool'>": "{Boolean}",
    "<class 'float'>": "{Number}",
    "typing.List[typing.Dict]": "{Array}",
    "typing.Dict": "{Object}",
    "typing.Optional[typing.Dict]": "{Object}",
    "typing.Dict[str, typing.List]": "{Object}",
    "typing.List[dict]": "{Array}",
    "typing.List[str]": "{Array}",
    "typing.List[int]": "{Array}",
    "typing.List[typing.List[dict]]": "{Array}",
    "typing.Union[int, bool]": "{Any}",
    "typing.Union[typing.Dict, NoneType]": "{Any}",
    "<class 'dict'>": "{Object}",
    "typing.Optional[int]": "{BigInteger|null}",
    "typing.Optional[str]": "{String|null}",
    "typing.Dict[str, typing.Dict]": "{Object}",
}

param_trans = {
    "False": "false",
    "True": "true",
    "None": "null",
    "0": "0",
    "2000": "2000",
    "title": '"title"',
}


def _add_stars(str_in: str) -> str:
    str_in = str_in or ""
    out = ""
    for line in str_in.split("\n")[1:]:
        out = out + "     * " + line[4:] + "\n"
    return "     " + out.strip()


class Command(BaseCommand):
    def handle(self, *args, **options):
        for key in sorted(all_methods.keys()):
            doc_block = _add_stars(all_methods.get(key).__doc__)
            sign = signature(all_methods.get(key))
            rpc_params = []
            sign_params = []
            for param in list(sign.parameters.items())[1:]:
                doc_block = (
                    doc_block
                    + "\n"
                    + PARAM_TPL.format(types_trans[str(param[1].annotation)], param[0])
                )
                rpc_params.append(param[0])
                if str(param[1].default) == "<class 'inspect._empty'>":
                    sign_params.append(param[0])
                else:
                    sign_params.append(
                        "{} = {}".format(param[0], param_trans[str(param[1].default)])
                    )
            doc_block = (
                doc_block
                + "\n"
                + RETURN_TPL.format(types_trans[str(sign.return_annotation)])
            )
            print(
                FN_TPL.format(
                    doc_block,
                    underscore_to_camel(key),
                    ", ".join(sign_params),
                    key,
                    ", ".join(rpc_params),
                )
            )



from django.core.management.base import BaseCommand
from django.conf import settings
import os
import time
import shutil


class Command(BaseCommand):
    def handle(self, *args, **options):
        now = time.time()
        max_age = 60 * 60 * 24
        sub_folders = [
            f.path for f in os.scandir(settings.PARTIAL_UPLOADS_DIR) if f.is_dir()
        ]
        for folder in sub_folders:
            folder_time = os.path.getmtime(folder)
            if now - folder_time > max_age:
                shutil.rmtree(folder)



# Generated by Django 3.1.1 on 2021-01-28 10:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0003_resource_ptr_owner"),
    ]

    operations = [
        migrations.CreateModel(
            name="MetadataType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128, verbose_name="titre")),
            ],
        ),
        migrations.AddField(
            model_name="metadata",
            name="data_type",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="resources.metadatatype",
            ),
        ),
    ]



# Generated by Django 3.2.2 on 2021-06-22 07:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0009_alter_metadatatype_title"),
    ]

    operations = [
        migrations.AddField(
            model_name="file",
            name="post_treated",
            field=models.BooleanField(default=True, verbose_name="traitements"),
        ),
    ]



from django.http import (
    HttpRequest,
    HttpResponseNotFound,
    HttpResponseBadRequest,
    FileResponse,
)
import pyvips
from resources.models import File
import os.path
from typing import Union
from jama import settings


def thumb(
    request: HttpRequest, sha256: str, angle: float, scale: float, crop: str
) -> Union[FileResponse, HttpResponseNotFound, HttpResponseBadRequest]:
    pic_file = File.objects.filter(hash=sha256).first()
    if not pic_file:
        return HttpResponseNotFound("Not Found")
    try:
        top, right, bottom, left = crop.split(",")
        top = int(top)
        right = int(right)
        bottom = int(bottom)
        left = int(left)
        angle = float(angle)
        scale = float(scale)
        if angle > 360 or angle < 0 or scale < 0 or scale > 1:
            raise ValueError
    except ValueError:
        return HttpResponseBadRequest("Bad Request")
    tmp_pic_path = "{}/thumb-{}-{}-{}-{}-{}-{}-{}.jpg".format(
        settings.TMP_THUMBNAILS_DIR, sha256, angle, scale, top, right, bottom, left
    )
    if not os.path.isfile(tmp_pic_path):
        crop_color = [96]
        pyvips.cache_set_max(0)
        img = pyvips.Image.new_from_file(pic_file.local_path())
        if angle:
            img = img.rotate(angle)
        img = img.draw_rect(crop_color, 0, 0, left, img.height, fill=True)
        img = img.draw_rect(crop_color, 0, 0, img.width, top, fill=True)
        img = img.draw_rect(
            crop_color, 0, img.height - bottom, img.width, bottom, fill=True
        )
        img = img.draw_rect(
            crop_color, img.width - right, 0, right, img.height, fill=True
        )
        img = img.resize(scale)
        img.write_to_file(tmp_pic_path)
    return FileResponse(open(tmp_pic_path, "rb"))



# Generated by Django 3.2.2 on 2021-09-21 08:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0017_auto_20210920_1727"),
    ]

    operations = [
        migrations.AlterField(
            model_name="project",
            name="label",
            field=models.TextField(unique=True),
        ),
    ]



# Generated by Django 3.1.1 on 2021-02-25 13:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("resources", "0006_collection_public_access"),
    ]

    operations = [
        migrations.CreateModel(
            name="Tag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uid", models.TextField()),
                ("ark", models.TextField(blank=True, null=True)),
                ("label", models.TextField(blank=True, null=True)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "étiquette",
                "verbose_name_plural": "étiquettes",
                "unique_together": {("uid", "owner")},
            },
        ),
        migrations.AddField(
            model_name="collection",
            name="tags",
            field=models.ManyToManyField(to="resources.Tag"),
        ),
        migrations.AddField(
            model_name="resource",
            name="tags",
            field=models.ManyToManyField(to="resources.Tag"),
        ),
    ]



# Generated by Django 5.1.5 on 2025-03-21 12:18

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0038_file_tiled"),
    ]

    operations = [
        migrations.AlterField(
            model_name="project",
            name="use_exiftool",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 3.1.1 on 2021-02-15 14:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0005_descendants_count"),
    ]

    operations = [
        migrations.AddField(
            model_name="collection",
            name="public_access",
            field=models.BooleanField(default=False, verbose_name="accès public"),
        ),
    ]



# noinspection PyUnresolvedReferences
from huey.contrib.djhuey import db_task
import logging
from typing import List
from django.conf import settings
from django.urls import reverse
from django.utils import timezone
from resources import models
from PIL import Image

logger = logging.getLogger(__name__)


@db_task()
def exif_task(file_id: int):
    from resources.models import File
    from resources.helpers import set_exif_metas

    try:
        f = File.objects.get(pk=file_id)
        set_exif_metas(f)
        if f.should_have_iiif():
            with Image.open(f.local_path()) as image:
                width, height = image.size
                f.denormalized_image_width = width
                f.denormalized_image_height = height
                f.save()

    except Exception as e:
        logger.warning("exit_task({}) failed".format(file_id))
        logger.info(repr(e))


@db_task()
def iiif_task(file_id: int):
    from resources.models import File
    from resources.helpers import make_iiif

    try:
        f = File.objects.get(pk=file_id)
        make_iiif(f)
    except Exception as e:
        logger.warning("iiif_task({}) failed".format(file_id))
        logger.info(repr(e))


@db_task()
def ocr_task(file_id: int):
    from resources.models import File
    from resources.helpers import make_ocr

    try:
        f = File.objects.get(pk=file_id)
        make_ocr(f)
    except Exception as e:
        logger.warning("ocr_task({}) failed".format(file_id))
        logger.info(repr(e))


@db_task()
def set_ark_to_resource(resource_id: int, location: str = None):
    if settings.ARK_SERVER and settings.ARK_APP_ID and settings.ARK_SECRET_KEY:
        from resources.models import Resource

        resource = Resource.objects.filter(pk=resource_id).first()
        if not resource:
            logger.warning(f"resource {resource_id} does not exist")
            return
        if resource.ark and not location:
            logger.info(f"resource {resource_id} has ark already, no update needed.")
            return
        import ark_client

        if not location:
            base_url = (
                settings.JAMA_SITE[:-1]
                if settings.JAMA_SITE[-1] == "/"
                else settings.JAMA_SITE
            )
            location = base_url + reverse(
                "ark_resource", kwargs={"resource_id": resource_id}
            )
        client = ark_client.Client(
            settings.ARK_APP_ID, settings.ARK_SECRET_KEY, settings.ARK_SERVER
        )
        if resource.ark:
            client.update(resource.ark, location)
        else:
            ark_name = client.create(location)
            resource.ark = ark_name
            resource.save()


@db_task()
def set_ark_to_collection(collection_id: int, location: str = None):
    if settings.ARK_SERVER and settings.ARK_APP_ID and settings.ARK_SECRET_KEY:
        from resources.models import Collection

        collection = Collection.objects.filter(pk=collection_id).first()
        if not collection:
            logger.warning(f"collection {collection_id} does not exist")
            return
        if collection.ark and not location:
            logger.info(
                f"collection {collection_id} has ark already, no update needed."
            )
            return
        import ark_client

        if not location:
            base_url = (
                settings.JAMA_SITE[:-1]
                if settings.JAMA_SITE[-1] == "/"
                else settings.JAMA_SITE
            )
            location = base_url + reverse(
                "ark_collection", kwargs={"collection_id": collection_id}
            )
        client = ark_client.Client(
            settings.ARK_APP_ID, settings.ARK_SECRET_KEY, settings.ARK_SERVER
        )
        if collection.ark:
            client.update(collection.ark, location)
        else:
            ark_name = client.create(location)
            collection.ark = ark_name
            collection.save()


@db_task()
def recursive_set_metas_to_collection(
    user_id: int, collection_id: int, metas: List[dict], user_task_id: int = None
) -> bool:
    r"""
    Sets all metas for a unique metadata set.

    Metas is a list of metadata id => metadata value dictionaries.

    All metas must share the same metadata set.

    the meta will be set to all direct children resources.

    /!\ *Not* actually recursive: Descendants (sub-collections and sub-collections resources) are IGNORED.
    """
    from rpc.methods import (
        remove_meta_value_from_collection,
        add_meta_to_collection,
        set_metas_to_resource,
    )
    from resources.models import Metadata, Collection, MetadataSet, UserTask
    from django.contrib.auth.models import User

    user_task = UserTask.objects.filter(pk=user_task_id).first()
    if user_task:
        user_task.started_at = timezone.now()
        user_task.save()

    # prevent mixing metas from different sets
    metadatasets_ids = []
    for meta_dict in metas:
        meta = Metadata.objects.get(pk=meta_dict["id"])
        metadatasets_ids.append(meta.set.pk)
    metadatasets_ids = list(set(metadatasets_ids))
    if len(metadatasets_ids) > 1:
        if user_task:
            user_task.failed_at = timezone.now()
            user_task.save()
        return False

    # check that objects exist
    collection_instance = Collection.objects.filter(
        pk=collection_id, deleted_at__isnull=True
    ).first()
    if not collection_instance:
        if user_task:
            user_task.failed_at = timezone.now()
            user_task.save()
        return False
    metadataset_instance = MetadataSet.objects.filter(pk=metadatasets_ids[0]).first()
    if not metadataset_instance:
        if user_task:
            user_task.failed_at = timezone.now()
            user_task.save()
        return False

    user = User.objects.filter(pk=user_id).first()
    if not user:
        if user_task:
            user_task.failed_at = timezone.now()
            user_task.save()
        return False

    # remove all metas from the set
    for meta_value in collection_instance.metadatacollectionvalue_set.filter(
        metadata__set=metadataset_instance
    ):
        remove_meta_value_from_collection(user, collection_id, meta_value.pk)

    # add metas
    for meta_dict in metas:
        add_meta_to_collection(user, collection_id, meta_dict["id"], meta_dict["value"])

    for resource in collection_instance.resources.filter(deleted_at__isnull=True):
        set_metas_to_resource(user, resource.pk, metas)

    if user_task:
        user_task.finished_at = timezone.now()
        user_task.save()
    return True


@db_task()
def update_resources_from_xlsx_rows(
    user_id: int, xlsx_rows: List[dict], user_task_id: int = None
):
    from django.contrib.auth.models import User
    from resources.models import UserTask

    project = None
    if len(xlsx_rows) > 0:
        first_resource = models.Resource.objects.filter(
            pk=xlsx_rows[0].get("pk")
        ).first()
        if first_resource:
            project = first_resource.ptr_project
    user_task = None
    user = User.objects.filter(pk=user_id).first()
    if user:
        if user_task_id is not None:
            user_task = UserTask.objects.filter(pk=user_task_id).first()
            if user_task:
                user_task.project = project
                user_task.started_at = timezone.now()
                user_task.save()
        from rpc.methods import update_resource_from_xlsx_row, ServiceException

        for xlsx_row in xlsx_rows:
            try:
                update_resource_from_xlsx_row(user, xlsx_row)
            except ServiceException as e:
                logger.warning(e)
        if user_task:
            user_task.finished_at = timezone.now()
            user_task.save()



# Generated by Django 5.0.1 on 2024-04-04 12:15

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0033_alter_usertask_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="usertask",
            name="project",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="resources.project",
            ),
        ),
    ]



from django.core.management.commands.shell import Command
from django.contrib.auth.models import User
from rich.console import Console
from rich.table import Table
from typing import List

# from pick import pick
from resources.models import (
    Resource,
    Collection,
    Project,
    APIKey,
    ProjectAccess,
    Permission,
    File,
)


def _console_table(title: str, headers: List["str"]) -> Table:
    table = Table(title=title)
    for header in headers:
        table.add_column(header)
    return table


def user(*args, **kwargs):
    return User.objects.get(*args, **kwargs)


def project(*args, **kwargs):
    return Project.objects.get(*args, **kwargs)


def projects(*args, **kwargs):
    table = _console_table(
        "Projects",
        [
            "pk",
            "label",
            "description",
            "admin mail",
            "ARK redirect",
            "exiftool",
            "users",
        ],
    )

    for p in Project.objects.filter(*args, **kwargs):
        table.add_row(
            str(p.pk),
            str(p.label),
            str(p.description),
            str(p.admin_mail),
            str(p.ark_redirect),
            str(p.use_exiftool),
            ", ".join(
                f"{a.user.username} ({a.user.pk})"
                for a in ProjectAccess.objects.filter(project=p)
            ),
        )
    Console().print(table)


def users(*args, **kwargs):
    # option, index = pick(["some", "of", "this"], "choose...", multiselect=True)
    table = _console_table(
        "Users", ["pk", "username", "email", "active", "superuser", "projects"]
    )
    for u in User.objects.filter(*args, **kwargs):
        table.add_row(
            str(u.pk),
            str(u.username),
            str(u.email),
            str(u.is_active),
            str(u.is_superuser),
            ", ".join(
                f"{a.project.label} ({a.project.pk})"
                for a in ProjectAccess.objects.filter(user=u)
            ),
        )
    Console().print(table)


def accesses():
    for a in ProjectAccess.objects.order_by("project"):
        print(a.pk, a)


def bpython(self, options):
    import bpython

    bpython.embed(
        {
            "APIKey": APIKey,
            "Collection": Collection,
            "File": File,
            "Permission": Permission,
            "Project": Project,
            "ProjectAccess": ProjectAccess,
            "Resource": Resource,
            "User": User,
            "projects": projects,
            "users": users,
            "accesses": accesses,
        }
    )


Command.bpython = bpython



# Generated by Django 3.1.2 on 2020-11-23 15:03

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Collection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=256, verbose_name="titre")),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="resources.collection",
                    ),
                ),
            ],
            options={
                "verbose_name": "collection",
                "verbose_name_plural": "collections",
            },
        ),
        migrations.CreateModel(
            name="CollectionMembership",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("rank", models.IntegerField(default=0, verbose_name="ordre")),
                (
                    "collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="resources.collection",
                    ),
                ),
            ],
            options={
                "verbose_name": "ressource <-> collection",
                "verbose_name_plural": "ressources <-> collections",
            },
        ),
        migrations.CreateModel(
            name="FileExtension",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        max_length=32, unique=True, verbose_name="extension"
                    ),
                ),
            ],
            options={
                "verbose_name": "extension de fichier",
                "verbose_name_plural": "extensions de fichier",
            },
        ),
        migrations.CreateModel(
            name="Metadata",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128, verbose_name="titre")),
                ("rank", models.IntegerField(default=0, verbose_name="ordre")),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "meta-donnée",
                "verbose_name_plural": "meta-données",
            },
        ),
        migrations.CreateModel(
            name="MetadataSet",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128, verbose_name="titre")),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Groupe de meta-données",
                "verbose_name_plural": "Groupes de meta-données",
                "unique_together": {("title", "owner")},
            },
        ),
        migrations.CreateModel(
            name="ResourceClass",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=64, verbose_name="titre")),
                (
                    "verbose_title",
                    models.CharField(max_length=128, verbose_name="titre long"),
                ),
                ("description", models.TextField(verbose_name="Description")),
                (
                    "set",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="resources.metadataset",
                    ),
                ),
            ],
            options={
                "verbose_name": "classe de ressource",
                "verbose_name_plural": "classes de ressources",
            },
        ),
        migrations.CreateModel(
            name="Resource",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128, verbose_name="titre")),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "collections",
                    models.ManyToManyField(
                        through="resources.CollectionMembership",
                        to="resources.Collection",
                    ),
                ),
                (
                    "resource_class",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="resources.resourceclass",
                    ),
                ),
            ],
            options={
                "verbose_name": "ressource",
                "verbose_name_plural": "ressources",
            },
        ),
        migrations.CreateModel(
            name="MetadataResourceValue",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.TextField()),
                ("sha1digest", models.CharField(max_length=40, unique=True)),
                (
                    "metadata",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="resources.metadata",
                    ),
                ),
                (
                    "resource",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="resources.resource",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="MetadataCollectionValue",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.TextField()),
                ("sha1digest", models.CharField(max_length=40, unique=True)),
                (
                    "collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="resources.collection",
                    ),
                ),
                (
                    "metadata",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="resources.metadata",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="metadata",
            name="set",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="resources.metadataset"
            ),
        ),
        migrations.CreateModel(
            name="FileType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=128, verbose_name="titre")),
                (
                    "mime",
                    models.CharField(
                        max_length=128, unique=True, verbose_name="type MIME"
                    ),
                ),
                (
                    "serve_with_iiif",
                    models.BooleanField(default=False, verbose_name="Servir par IIIF"),
                ),
                ("extensions", models.ManyToManyField(to="resources.FileExtension")),
            ],
            options={
                "verbose_name": "type de fichier",
                "verbose_name_plural": "types de fichier",
            },
        ),
        migrations.AddField(
            model_name="collectionmembership",
            name="resource",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="resources.resource"
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="resources",
            field=models.ManyToManyField(
                through="resources.CollectionMembership", to="resources.Resource"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="metadata",
            unique_together={("title", "set", "owner")},
        ),
        migrations.AlterUniqueTogether(
            name="collectionmembership",
            unique_together={("resource", "collection")},
        ),
        migrations.AlterUniqueTogether(
            name="collection",
            unique_together={("title", "owner", "parent")},
        ),
        migrations.CreateModel(
            name="APIKey",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "key",
                    models.CharField(max_length=40, unique=True, verbose_name="clef"),
                ),
                ("active", models.BooleanField(default=False, verbose_name="actif")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "clef d'API",
                "verbose_name_plural": "clefs d'API",
                "unique_together": {("key", "user")},
            },
        ),
        migrations.CreateModel(
            name="File",
            fields=[
                (
                    "resource_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="resources.resource",
                    ),
                ),
                (
                    "original_name",
                    models.CharField(max_length=256, verbose_name="nom d'origine"),
                ),
                ("hash", models.CharField(max_length=64, verbose_name="hachage")),
                ("size", models.BigIntegerField(verbose_name="taille")),
                (
                    "file_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.RESTRICT,
                        to="resources.filetype",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "fichier",
                "verbose_name_plural": "fichiers",
                "unique_together": {("owner", "hash")},
            },
            bases=("resources.resource",),
        ),
    ]



from django.test import TestCase
from resources import models
from rpc import methods as rpc_methods
from resources.management.commands.loadfixtures import load_fixtures
from django.contrib.auth.models import User
from annotations.models import Annotation
from rpc.methods import ServiceException

test_json = {"pouet": "tagada"}


class AnnotationTestCase(TestCase):
    def setUp(self):
        # load file types, create permissions
        load_fixtures()

        self.test_user = User.objects.create(username="basic_user")
        self.basic_user_key = "acme"

        # only django superusers have access to certain rpc methods.
        self.admin_user = User.objects.create(username="admin_user", is_superuser=True)
        rpc_methods.activate_rpc_access(
            self.admin_user, self.test_user.username, self.basic_user_key
        )

        self.test_project = models.Project.objects.create(
            label="projet test", description="projet test"
        )
        self.test_project_root_collection = models.Collection.objects.create(
            title="root", parent_id=None, project=self.test_project
        )
        self.test_metadataset = models.MetadataSet.objects.create(
            title="test metadata set", project=self.test_project
        )

        self.admin_role = models.Role.objects.create(
            label="admin", project=self.test_project
        )
        # give all permissions to admin role
        for perm in models.Permission.objects.all():
            self.admin_role.permissions.add(perm)

        # give admin role to the user on the project
        models.ProjectAccess.objects.get_or_create(
            project=self.test_project, user=self.test_user, role=self.admin_role
        )
        models.ProjectAccess.objects.get_or_create(
            project=self.test_project, user=self.admin_user, role=self.admin_role
        )
        # public user
        self.public_user = User.objects.create(username="public_user")
        self.public_role = models.Role.objects.create(
            label="public", project=self.test_project
        )
        for perm in models.Permission.objects.all():
            self.public_role.permissions.add(perm)
        models.ProjectAccess.objects.get_or_create(
            project=self.test_project, user=self.public_user, role=self.public_role
        )

    def test_add_annotation(self):
        with self.assertRaises(ServiceException):
            rpc_methods.create_annotation(self.test_user, 42, test_json)

        res = models.Resource.objects.create(
            title="test res", ptr_project=self.test_project
        )
        serialized_annotation = rpc_methods.create_annotation(
            self.test_user, res.pk, test_json
        )
        self.assertEqual(serialized_annotation.get("data"), test_json)

    def test_delete_annotation(self):
        with self.assertRaises(ServiceException):
            rpc_methods.delete_annotation(self.test_user, 42)
        res = models.Resource.objects.create(
            title="test res", ptr_project=self.test_project
        )
        serialized_annotation = rpc_methods.create_annotation(
            self.test_user, res.pk, test_json
        )
        # not owner
        with self.assertRaises(ServiceException):
            rpc_methods.delete_annotation(
                self.admin_user, serialized_annotation.get("id")
            )
        self.assertTrue(
            rpc_methods.delete_annotation(
                self.test_user, serialized_annotation.get("id")
            )
        )

    def test_list_annotations(self):
        res = models.Resource.objects.create(
            title="test res", ptr_project=self.test_project
        )
        rpc_methods.create_annotation(self.test_user, res.pk, test_json)
        rpc_methods.create_annotation(self.test_user, res.pk, test_json)
        list = rpc_methods.list_annotations(self.test_user, res.pk)
        self.assertEqual(len(list), 2)

    def test_update_annotation(self):
        res = models.Resource.objects.create(
            title="test res", ptr_project=self.test_project
        )
        serialized_annotation = rpc_methods.create_annotation(
            self.test_user, res.pk, test_json
        )
        rpc_methods.update_annotation(
            self.test_user, serialized_annotation.get("id"), {"something": "else"}
        )
        fetched_annotation = Annotation.objects.get(pk=serialized_annotation.get("id"))
        self.assertEqual(fetched_annotation.data, {"something": "else"})



# Generated by Django 3.2.2 on 2021-09-22 12:25

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0019_alter_collection_project"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="admin_mail",
            field=models.EmailField(max_length=254, null=True),
        ),
    ]



# Generated by Django 3.2.2 on 2021-09-20 14:19

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0014_auto_20210915_2024"),
    ]

    operations = [
        migrations.AddField(
            model_name="metadataset",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="metadataset",
            name="updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# Generated by Django 3.1.1 on 2021-02-10 05:59
from django.conf import settings
from django.db import migrations

if settings.JAMA_SQLITE_DB_PATH:
    CREATE_SQL = ""
    DROP_SQL = ""
else:
    CREATE_SQL = """
    CREATE OR REPLACE FUNCTION get_all_collection_descendants_array(use_parent INT4) RETURNS INT4[] as $$
    DECLARE
        process_parents INT4[] := ARRAY[ use_parent ];
        children INT4[] := '{}';
        new_children INT4[];
    BEGIN
        WHILE ( array_upper( process_parents, 1 ) IS NOT NULL ) LOOP
            new_children := ARRAY( SELECT id FROM resources_collection WHERE parent_id = ANY( process_parents ) AND id <> ALL( children ) );
            children := children || new_children;
            process_parents := new_children;
        END LOOP;
        RETURN children;
    END;
    $$ LANGUAGE plpgsql;
    """

    DROP_SQL = """
    DROP FUNCTION IF EXISTS get_all_collection_descendants_array(INT4);
    """


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0004_auto_20210128_1046"),
    ]

    operations = [migrations.RunSQL(CREATE_SQL, DROP_SQL)]



from resources import models
from django.contrib.auth.models import User


class SerializerCache:
    def __init__(self):
        self._cache_dict = {}

    def fetch_obj(self, obj_type: str, obj_id: int):
        cache_key = "{}{}".format(obj_type, obj_id)
        if cache_key not in self._cache_dict.keys():
            if obj_type == "metadata":
                self._cache_dict[cache_key] = models.Metadata.objects.get(pk=obj_id)
            elif obj_type == "metadataset":
                self._cache_dict[cache_key] = models.MetadataSet.objects.get(pk=obj_id)
            elif obj_type == "user":
                self._cache_dict[cache_key] = User.objects.get(pk=obj_id)
            elif obj_type == "project":
                self._cache_dict[cache_key] = models.Project.objects.get(pk=obj_id)
            elif obj_type == "role":
                self._cache_dict[cache_key] = models.Role.objects.get(pk=obj_id)
            elif obj_type == "file_type":
                self._cache_dict[cache_key] = models.FileType.objects.get(pk=obj_id)
            else:
                raise ValueError("unsupported object type: {}".format(obj_type))
        return self._cache_dict.get(cache_key)



# Generated by Django 4.2.5 on 2023-10-04 12:33

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0028_collection_ark_resource_ark"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="ark_redirect",
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 3.2.2 on 2021-08-30 03:11

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0012_auto_20210713_0807"),
    ]

    operations = [
        migrations.AddField(
            model_name="collection",
            name="is_oai_record",
            field=models.BooleanField(default=False, verbose_name="enregistrement OAI"),
        )
    ]



# Generated by Django 3.2.2 on 2021-07-13 08:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0011_collection_representative"),
    ]

    operations = [
        migrations.AddField(
            model_name="collection",
            name="deleted_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="resource",
            name="deleted_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.0 on 2022-01-21 10:23

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0024_alter_metadata_options"),
    ]

    operations = [
        migrations.CreateModel(
            name="ProjectProperty",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=64)),
                ("value", models.JSONField()),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="resources.project",
                    ),
                ),
            ],
            options={
                "unique_together": {("project", "key")},
            },
        ),
    ]



from django.utils.text import slugify
from django.db import models, connection
from django.contrib.auth.models import User
import os
import json
from jama import settings
import hashlib
from typing import Iterator, Union, List
from django.db.models.signals import post_save
from django.dispatch import receiver
from subprocess import call
from django.utils.timezone import now
from resources import tasks
import pytesseract
from PIL import Image, UnidentifiedImageError

Image.MAX_IMAGE_PIXELS = None

XLSX_MULTIPLE_VALUES_SEPARATOR = "\n--\n"


def _flatten_resource(
    resource: "Resource", known_metadatas: dict, metadatas_labels: List
) -> dict:
    data = {"pk": resource.pk, "title": resource.title}
    metadata_ids = known_metadatas.keys()
    for known_metadata_label in metadatas_labels:
        data[known_metadata_label] = None
    for mr in (
        resource.metadataresourcevalue_set.exclude(metadata__title="ExifTool")
        .exclude(metadata__title="OCR")
        .exclude(metadata__title="scd_cms")
        .select_related("metadata")
    ):
        if mr.metadata.pk in metadata_ids:
            meta_key = f"{str(mr.metadata)}"
            if data[meta_key]:
                data[meta_key] = (
                    data[meta_key] + XLSX_MULTIPLE_VALUES_SEPARATOR + mr.value
                )
            else:
                data[meta_key] = mr.value
    return data


def hash_to_iiif_path(file_hash: str, separator: str = os.path.sep) -> str:
    return "{}{}{}{}{}".format(
        file_hash[:2], separator, file_hash[2:4], separator, file_hash
    )


def hash_to_local_path(file_hash: str) -> str:
    return "{}{}{}{}{}{}{}".format(
        settings.MEDIA_FILES_DIR.rstrip("/"),
        os.path.sep,
        file_hash[:2],
        os.path.sep,
        file_hash[2:4],
        os.path.sep,
        file_hash,
    )


def hash_to_iiif_thumbnail_url(
    file_hash: str, width: int = None, height: int = None
) -> str:
    return "{}{}/full/{}{},{}/0/default.jpg".format(
        settings.JAMA_IIIF_ENDPOINT,
        hash_to_iiif_path(file_hash, settings.IIIF_PATH_SEPARATOR),
        settings.JAMA_IIIF_UPSCALING_PREFIX,
        width or "",
        height or "",
    )


class NotIIIF(RuntimeError):
    pass


class APIKey(models.Model):
    key = models.CharField("clef", max_length=40, unique=True)
    active = models.BooleanField("actif", default=False)
    user = models.ForeignKey(User, on_delete=models.CASCADE)

    def __str__(self):
        return "{}: {}".format(self.user, self.key)

    class Meta:
        verbose_name = "clef d'API"
        verbose_name_plural = "clefs d'API"
        unique_together = (("key", "user"),)


class Project(models.Model):
    """
    Project is used for sharing collections
    and resources between users.
    """

    label = models.TextField(unique=True)
    description = models.TextField()
    admin_mail = models.EmailField(null=True)
    #
    # Intended as a redirect pattern for ark resolving.
    # Example: https://service.tld/ark/redirect/[CLASS]/[ARK].
    #
    # Replaced values are:
    # - [CLASS]: "resource" or "collection"
    # - [ARK]: ARK name (Resource.ark or Collection.ark)
    # - [PK]: Resource.pk or Collection.pk
    #
    # See jama.views.ark_resource and jama.views.ark_collection.
    #
    ark_redirect = models.TextField(null=True)
    # Exiftool returns a lot of data and you may not need it in your
    # project. When set to false, the exiftool task is bypassed.
    use_exiftool = models.BooleanField(default=False)

    def __str__(self):
        return self.label

    @property
    def root_collection(self) -> "Collection":
        try:
            col, created = Collection.objects.get_or_create(project=self, parent=None)
            if created:
                col.title = "root {}".format(self.label)
                col.save()
            return col
        except Collection.MultipleObjectsReturned:
            return Collection.objects.filter(project=self, parent=None).first()

    def metadatas(self, exclude_automatic_metas=True) -> List["Metadata"]:
        metadatas = []
        query = MetadataSet.objects.filter(project=self).order_by("title")
        if exclude_automatic_metas:
            query = query.exclude(title="ExifTool").exclude(title="OCR")
        for metadataset in query:
            for metadata in Metadata.objects.filter(set=metadataset).order_by("title"):
                metadatas.append(metadata)
        return metadatas


class ProjectProperty(models.Model):
    """
    Generic key/value store for the project.

    Store anything that is needed by the client application,
    like user prefs, field labels or other useful data.

    Be extra careful with access rights here.
    """

    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    key = models.CharField(max_length=64, null=False, blank=False)
    value = models.JSONField()

    class Meta:
        unique_together = (("project", "key"),)


class UserTask(models.Model):
    owner = models.ForeignKey(User, on_delete=models.CASCADE)
    description = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    started_at = models.DateTimeField(null=True)
    finished_at = models.DateTimeField(null=True)
    failed_at = models.DateTimeField(null=True)
    canceled_at = models.DateTimeField(null=True)
    project = models.ForeignKey(Project, null=True, on_delete=models.CASCADE)

    class Meta:
        ordering = ["-updated_at"]


class Tag(models.Model):
    uid = models.TextField()
    ark = models.TextField(null=True, blank=True)
    label = models.TextField(null=True, blank=True)  # pref label de SKOS
    project = models.ForeignKey(Project, on_delete=models.CASCADE)

    def __str__(self):
        return self.label or self.uid

    class Meta:
        verbose_name = "étiquette"
        verbose_name_plural = "étiquettes"
        unique_together = (("uid", "project"),)


class FileExtension(models.Model):
    label = models.CharField("extension", max_length=32, unique=True)

    def __str__(self):
        return ".{}".format(self.label)

    class Meta:
        verbose_name = "extension de fichier"
        verbose_name_plural = "extensions de fichier"


class FileType(models.Model):
    title = models.CharField("titre", max_length=128)
    mime = models.CharField("type MIME", max_length=128, unique=True)
    extensions = models.ManyToManyField(FileExtension)
    serve_with_iiif = models.BooleanField("Servir par IIIF", default=False)

    def __str__(self):
        return "{}".format(self.title)

    class Meta:
        verbose_name = "type de fichier"
        verbose_name_plural = "types de fichier"


class MetadataSet(models.Model):
    title = models.CharField("titre", max_length=128)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "{}".format(self.title)

    class Meta:
        verbose_name = "Groupe de meta-données"
        verbose_name_plural = "Groupes de meta-données"
        unique_together = (("title", "project"),)


class MetadataType(models.Model):
    title = models.CharField("titre", max_length=128, unique=True)

    def __str__(self):
        return self.title


class Metadata(models.Model):
    title = models.CharField("titre", max_length=128)
    set = models.ForeignKey(MetadataSet, on_delete=models.CASCADE)
    rank = models.IntegerField("ordre", default=0)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    data_type = models.ForeignKey(MetadataType, null=True, on_delete=models.SET_NULL)
    expose = models.BooleanField(default=True)

    def __str__(self):
        return "{}: {}".format(self.set.title, self.title)

    class Meta:
        verbose_name = "meta-donnée"
        verbose_name_plural = "meta-données"
        unique_together = (("title", "set", "project"),)
        ordering = ["rank", "title"]


class ResourceClass(models.Model):
    title = models.CharField("titre", max_length=64)
    verbose_title = models.CharField("titre long", max_length=128)
    description = models.TextField("Description")
    set = models.ForeignKey(MetadataSet, null=True, on_delete=models.SET_NULL)

    def __str__(self):
        return "{}: {}".format(self.set.title, self.title)

    class Meta:
        verbose_name = "classe de ressource"
        verbose_name_plural = "classes de ressources"


class Resource(models.Model):
    title = models.CharField("titre", max_length=128)
    resource_class = models.ForeignKey(
        ResourceClass, null=True, on_delete=models.SET_NULL, blank=True
    )
    collections = models.ManyToManyField("Collection", through="CollectionMembership")
    tags = models.ManyToManyField(Tag)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    deleted_at = models.DateTimeField(null=True, blank=True)
    # this is duplicated as "project" at the File class level
    # to enforce a unique constraint on Project and File hash.
    # "ptr" stands for "pointer" since the Resource class should
    # not be used directly.
    ptr_project = models.ForeignKey(Project, on_delete=models.CASCADE, null=True)
    ark = models.CharField("nom ARK", max_length=64, null=True)

    def __str__(self):
        return "{}".format(self.title)

    class Meta:
        verbose_name = "ressource"
        verbose_name_plural = "ressources"
        ordering = ("title",)

    def soft_delete(self):
        self.deleted_at = now()
        self.save()


class File(Resource):
    original_name = models.CharField("nom d'origine", max_length=256)
    # this duplicates "ptr_project" from Resource class
    # (can't have unique constraints across tables unfortunately)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    hash = models.CharField("hachage", max_length=64)
    file_type = models.ForeignKey(FileType, on_delete=models.RESTRICT)
    size = models.BigIntegerField("taille")
    # shortcuts
    denormalized_image_width = models.IntegerField(
        "largeur de l'image", blank=True, null=True
    )
    denormalized_image_height = models.IntegerField(
        "hauteur de l'image", blank=True, null=True
    )
    text_boxes = models.JSONField("text boxes de tesseract", blank=True, null=True)
    tiled = models.BooleanField(default=False)

    def __str__(self):
        return "{}".format(self.title)

    def should_have_iiif(self) -> bool:
        return self.file_type.serve_with_iiif

    def image_width(self) -> Union[int, None]:
        if self.denormalized_image_width is None:
            self.denormalized_image_width = self.image_width_from_metas()
        return self.denormalized_image_width

    def image_height(self) -> Union[int, None]:
        if self.denormalized_image_height is None:
            self.denormalized_image_height = self.image_height_from_metas()
        return self.denormalized_image_height

    def image_width_from_metas(self) -> Union[int, None]:
        val = None
        if self.should_have_iiif():
            if self.project.use_exiftool:
                prop = self.metadataresourcevalue_set.filter(
                    metadata__title__iendswith="imagewidth",
                    metadata__set__title__iexact="exiftool",
                ).first()
                if prop:
                    try:
                        val = int(prop.value)
                    except ValueError:
                        pass
            else:
                image = Image.open(self.local_path())
                val, _ = image.size
        return val

    def image_height_from_metas(self) -> Union[int, None]:
        val = None
        if self.should_have_iiif():
            if self.project.use_exiftool:
                prop = self.metadataresourcevalue_set.filter(
                    metadata__title__iendswith="imageheight",
                    metadata__set__title__iexact="exiftool",
                ).first()
                if prop:
                    try:
                        val = int(prop.value)
                    except ValueError:
                        pass
            else:
                image = Image.open(self.local_path())
                _, val = image.size
        return val

    class Meta:
        verbose_name = "fichier"
        verbose_name_plural = "fichiers"
        unique_together = (("project", "hash"),)

    def filter_meta(self, full_meta_title: str, default_value=None):
        """
        Simple access to wanted meta. ":" is used as a separator.

        Ex:
            some_file.filter_meta("ExifTool:ID3:Album", "")
        """
        full_meta_title = full_meta_title.lower()
        for metaval in self.metadataresourcevalue_set.all():
            current_meta_title = "{}:{}".format(
                metaval.metadata.set.title, metaval.metadata.title
            ).lower()
            if current_meta_title == full_meta_title:
                return metaval.value
        return default_value

    def iiif_infos_url(self) -> str:
        if self.should_have_iiif():
            return "{}{}/info.json".format(
                settings.JAMA_IIIF_ENDPOINT,
                hash_to_iiif_path(self.hash, settings.IIIF_PATH_SEPARATOR),
            )
        else:
            raise NotIIIF("Resource is not served via IIIF")

    def iiif_thumbnail_url(self, width: int = None, height: int = None) -> str:
        if self.should_have_iiif():
            return "{}{}/full/{}{},{}/0/default.jpg".format(
                settings.JAMA_IIIF_ENDPOINT,
                hash_to_iiif_path(self.hash, settings.IIIF_PATH_SEPARATOR),
                settings.JAMA_IIIF_UPSCALING_PREFIX,
                width or "",
                height or "",
            )
        else:
            raise NotIIIF("Resource is not served via IIIF")

    def local_path(self):
        return hash_to_local_path(self.hash)

    def ocr_output(self) -> Union[None, str]:
        """
        Fetch OCR output if available. None if not available.
        """
        metas_set, created = MetadataSet.objects.get_or_create(
            title="OCR", project=self.project
        )
        meta, created = Metadata.objects.get_or_create(
            title="tesseract output", set=metas_set, project=self.project
        )
        try:
            meta_value = MetadataResourceValue.objects.get(metadata=meta, resource=self)
            return meta_value.value
        except MetadataResourceValue.DoesNotExist:
            return None

    def make_ocr(self, refresh: bool = False):
        """
        Performs OCR on file if OCR output not yet available.
        Use refresh = True to force OCR.
        """
        if not self.should_have_iiif():  # not a picture
            return
        if self.ocr_output() and not refresh:
            return

        try:
            im = Image.open(self.local_path())
            content = pytesseract.image_to_string(im)
            ocr_metas_set, _ = MetadataSet.objects.get_or_create(
                title="OCR", project=self.project
            )
            tesseract_meta, _ = Metadata.objects.get_or_create(
                project=self.project, title="tesseract output", set=ocr_metas_set
            )
            MetadataResourceValue.objects.filter(
                metadata=tesseract_meta, resource=self
            ).delete()
            MetadataResourceValue.objects.get_or_create(
                metadata=tesseract_meta, resource=self, value=content.strip()
            )
            boxes = pytesseract.image_to_boxes(im, output_type=pytesseract.Output.DICT)
            self.text_boxes = boxes
            self.save()
        except UnidentifiedImageError:
            pass

        if self.has_extension("pdf") or self.has_extension("ai"):
            # extract PDF text layer if available
            pdftotext_return_code = call(
                [
                    "pdftotext",
                    self.local_path(),
                    self.local_path() + ".pdftotext.txt",
                ]
            )
            if pdftotext_return_code == 0:
                with open(self.local_path() + ".pdftotext.txt", "r") as text_layer:
                    content = text_layer.read()
                    metas_set, _ = MetadataSet.objects.get_or_create(
                        title="OCR", project=self.project
                    )
                    meta, _ = Metadata.objects.get_or_create(
                        project=self.project, title="pdftotext output", set=metas_set
                    )
                    MetadataResourceValue.objects.filter(
                        metadata=meta, resource=self
                    ).delete()
                    MetadataResourceValue.objects.get_or_create(
                        metadata=meta, resource=self, value=content.strip()
                    )

    def has_extension(self, extension: str) -> bool:
        extension = extension.lower()
        for ext in self.file_type.extensions.all():
            if ext.label == extension:
                return True
        return False

    def save(self, *args, **kwargs):
        self.ptr_project = self.project
        super(File, self).save(*args, **kwargs)

    @property
    def new_filename(self) -> str:
        """
        Get new filename computed from title.
        Tries to keep original extension.
        """
        _, extension = os.path.splitext(self.original_name)
        if not extension:
            extension = "." + self.file_type.extensions.first().label
        base_name, _ = os.path.splitext(self.title)
        slugged_title = slugify(base_name)
        return slugged_title + extension.lower()

    @property
    def original_name_extension(self) -> str:
        _, ext = os.path.splitext(self.original_name)
        return ext


class Collection(models.Model):
    title = models.CharField("titre", max_length=256)
    resources = models.ManyToManyField(Resource, through="CollectionMembership")
    tags = models.ManyToManyField(Tag)
    project = models.ForeignKey(Project, on_delete=models.CASCADE, null=True)
    parent = models.ForeignKey("self", null=True, on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    deleted_at = models.DateTimeField(null=True, blank=True)
    public_access = models.BooleanField("accès public", default=False)
    # Used to mark a resource as representative of a collection.
    # Typical use case: set a miniature for a collection
    representative = models.ForeignKey(
        "Resource",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="represents",
    )
    ark = models.CharField("nom ARK", max_length=64, null=True)

    def __str__(self):
        return "{}".format(self.title)

    def ark_destination_url(self) -> str:
        if self.ark and self.project.ark_redirect:
            return self.project.ark_redirect.replace("[CLASS]", "collection").replace(
                "[PK]", self.pk
            )
        return ""

    def ancestors(self) -> List["Collection"]:
        ancestors = []
        parent = self.parent
        while parent:
            ancestors.append(parent)
            parent = parent.parent
        ancestors.reverse()
        return ancestors

    def children(self):
        return Collection.objects.filter(parent=self, deleted_at__isnull=True)

    def descendants(self) -> Iterator["Collection"]:
        yield from _recurse_collection(self)

    def descendants_resources(self) -> Iterator[Resource]:
        for res in self.resources.filter(deleted_at__isnull=True):
            yield res
        for col in _recurse_collection(self):
            for res in col.resources.filter(deleted_at__isnull=True):
                yield res

    def descendants_and_self_ids(self) -> List[int]:
        ids = [self.pk]
        for descendant in self.descendants():
            ids.append(descendant.pk)
        return ids

    def descendants_count(self) -> int:
        if settings.JAMA_SQLITE_DB_PATH:
            return sum(1 for _ in self.descendants())
        else:
            with connection.cursor() as cursor:
                cursor.execute(
                    """select count(*)
                    from resources_collection
                    where id = any(get_all_collection_descendants_array(%s))
                    and deleted_at is null;""",
                    [self.id],
                )
                row = cursor.fetchone()
                return int(row[0])

    def descendants_resources_count(self) -> int:
        """
        This counts resources from all the descendant collections,
        EXCLUDING the current collection's direct resources.
        """
        if settings.JAMA_SQLITE_DB_PATH:
            total = 0
            for col in self.descendants():
                total = total + col.resources.filter(deleted_at__isnull=True).count()
            return total
        else:
            with connection.cursor() as cursor:
                # For some reason, Postgresql will perform a fast index scan when given the list of collection ids as a
                # string literal in an ANY clause.
                # When given directly the procedure's result (list of int values), it will perform a slow sequential scan.
                # Solution: Prepare a string literal.

                # FIRST, get all descendants ids
                cursor.execute(
                    """select array_agg(id)
                from resources_collection
                where id = any(get_all_collection_descendants_array(%s))
                and deleted_at is null;""",
                    [self.id],
                )
                row = cursor.fetchone()
                if not row[0]:
                    return 0
                cursor.execute(
                    # Slow:
                    # "select count(*) from resources_collectionmembership where collection_id = any(get_all_collection_descendants_array(%s));",
                    # [self.id],
                    # Fast:
                    """select count(*) from resources_collectionmembership
                    inner join resources_resource on resources_collectionmembership.resource_id = resources_resource.id
                    where collection_id = any({})
                    and deleted_at is null;""".format(
                        # THEN, make the string literal.
                        # (TODO: write a procedure so PGSQL can make the literal string by itself)
                        "'{" + ",".join(str(x) for x in row[0]) + "}'"
                    ),
                )
                row = cursor.fetchone()
                return int(row[0])

    def descendant_resources_count(self) -> int:
        """
        This counts descendant resources INCLUDING the current collection's direct resources.
        """
        return (
            self.descendants_resources_count()
            + self.resources.filter(deleted_at__isnull=True).count()
        )

    class Meta:
        verbose_name = "collection"
        verbose_name_plural = "collections"
        unique_together = (("title", "project", "parent"),)
        ordering = ("title",)

    def iiif_manifest(self) -> str:
        manifest = {
            "@context": "http://iiif.io/api/presentation/3/context.json",
            "id": "{}collection/{}/manifest.json".format(settings.JAMA_SITE, self.id),
            "type": "Manifest",
            "label": {"en": [self.title]},
            "sequences": [
                {
                    "@id": "{}collection/{}/seq/0".format(settings.JAMA_SITE, self.id),
                    "@type": "sc:Sequence",
                    "canvases": [],
                }
            ],
        }
        query_set = File.objects.filter(
            collections__id=self.id,
            file_type__serve_with_iiif=True,
        ).order_by("collection__collectionmembership__rank")
        p = 0
        for image in query_set:
            manifest["sequences"][0]["canvases"].append(
                {
                    "@id": "https://acme.tld/someid",
                    "@type": "sc:Canvas",
                    "label": "Canvas",
                    "width": image.image_width(),
                    "height": image.image_height(),
                    "images": [
                        {
                            "@type": "oa:Annotation",
                            "motivation": "sc:painting",
                            "resource": {
                                "@id": image.iiif_infos_url(),
                                "@type": "dctypes:Image",
                                "format": "image/jpg",
                            },
                        }
                    ],
                }
            )
            p = p + 1
        return json.dumps(manifest, indent=4)

    def soft_delete(self):
        self.deleted_at = now()
        self.save()

    def dublin_core_metas(self) -> List["MetadataCollectionValue"]:
        metas = []
        try:
            dublin_core_set = MetadataSet.objects.get(
                title__iexact="Dublin Core", project=self.project
            )
            for metadata in self.metadatacollectionvalue_set.filter(
                metadata__set=dublin_core_set
            ):
                metas.append(metadata)
            return metas
        except MetadataSet.DoesNotExist:
            return []

    def dublin_core_title(self) -> str:
        dc_metas = self.dublin_core_metas()
        for m in dc_metas:
            if m.metadata.title == "title":
                return m.value
        return self.title

    def oai_setspec(self) -> str:
        """
        Get setSpec value for OAI-PMH xml.
        (see https://www.openarchives.org/OAI/openarchivesprotocol.html#Set)
        """
        ids = []
        ancestors = self.ancestors()
        for ancestor in ancestors[1:]:
            if ancestor.public_access:
                ids.append(str(ancestor.pk))
        ids.append(str(self.pk))
        return ":".join(ids)

    def to_path(self):
        titles = []
        for ancestor in self.ancestors():
            titles.append(ancestor.title)
        titles.append(self.title)
        return titles

    def yield_resource_data_for_export(self) -> Iterator[dict]:
        known_metadatas = {}
        known_metadatas_labels = []
        for metadata in self.project.metadatas():
            known_metadatas[metadata.pk] = metadata
            known_metadatas_labels.append(f"{str(metadata)}")
        for res in (
            self.resources.filter(deleted_at__isnull=True)
            .order_by("title")
            .prefetch_related("metadataresourcevalue_set")
        ):
            yield _flatten_resource(res, known_metadatas, known_metadatas_labels)
        for res in self.descendants_resources():
            yield _flatten_resource(res, known_metadatas, known_metadatas_labels)


def _recurse_collection(collection: Collection) -> Iterator[Collection]:
    try:
        for child in collection.children():
            yield child
            yield from _recurse_collection(child)
    except RecursionError:
        pass


class MetadataCollectionValue(models.Model):
    metadata = models.ForeignKey(Metadata, on_delete=models.CASCADE)
    collection = models.ForeignKey(Collection, on_delete=models.CASCADE)
    value = models.TextField()
    sha1digest = models.CharField(max_length=40, null=False, blank=False, unique=True)

    def save(self, *args, **kwargs):
        self.sha1digest = hashlib.sha1(
            "{}-{}-{}".format(self.metadata_id, self.collection_id, self.value).encode()
        ).hexdigest()
        super(MetadataCollectionValue, self).save(*args, **kwargs)


class MetadataResourceValue(models.Model):
    metadata = models.ForeignKey(Metadata, on_delete=models.CASCADE)
    resource = models.ForeignKey(Resource, on_delete=models.CASCADE)
    value = models.TextField()
    sha1digest = models.CharField(max_length=40, null=False, blank=False, unique=True)

    def save(self, *args, **kwargs):
        self.sha1digest = hashlib.sha1(
            "{}-{}-{}".format(self.metadata_id, self.resource_id, self.value).encode()
        ).hexdigest()
        super(MetadataResourceValue, self).save(*args, **kwargs)


class CollectionMembership(models.Model):
    resource = models.ForeignKey(Resource, on_delete=models.CASCADE)
    collection = models.ForeignKey(Collection, on_delete=models.CASCADE)
    rank = models.IntegerField("ordre", default=0)

    def __str__(self):
        return "{} <-> {}".format(self.collection.title, self.resource.title)

    class Meta:
        verbose_name = "ressource <-> collection"
        verbose_name_plural = "ressources <-> collections"
        unique_together = (("resource", "collection"),)


class Permission(models.Model):
    """
    Format labels like so: [object type].[action]

    Permissions are NOT project-specific, they are
    tied to the API.

    Examples:
        - collection.delete
        - resource.add
        - meta.modify
        - resource.move
    and so on...
    """

    label = models.TextField(unique=True)

    def __str__(self):
        return self.label


class Role(models.Model):
    label = models.TextField()
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    permissions = models.ManyToManyField(Permission)

    def __str__(self):
        return self.label

    class Meta:
        unique_together = (("label", "project"),)


class ProjectAccess(models.Model):
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    role = models.ForeignKey(Role, on_delete=models.CASCADE)

    def __str__(self):
        return "{}: {} ({})".format(self.project, self.user, self.role)

    class Meta:
        unique_together = (("project", "user", "role"),)


@receiver(post_save, sender=Resource)
def resource_post_save(sender, instance: Resource, **kwargs):
    if not instance.ark:
        tasks.set_ark_to_resource(instance.pk)


@receiver(post_save, sender=Collection)
def collection_post_save(sender, instance: Collection, **kwargs):
    tasks.set_ark_to_collection(instance.pk)



from django.core.management.base import BaseCommand
from resources.models import (
    FileExtension,
    FileType,
    MetadataSet,
    Metadata,
    MetadataType,
    ResourceClass,
    Permission,
    Project,
)
import json
import csv
import pathlib

IIIF_SUPPORT = [
    "application/pdf",
    "image/png",
    "image/jpg",
    "image/jpeg",
    "image/bmp",
    "image/gif",
    "image/tiff",
]


def set_base_permissions():
    obj_types = [
        "collection",
        "resource",
        "metadata",
        "metadataset",
        "file",
        "tag",
        "role",
        "permission",
        "project",
    ]
    permissions = ["create", "read", "update", "delete"]
    for obj_type in obj_types:
        for permission in permissions:
            Permission.objects.get_or_create(label="{}.{}".format(obj_type, permission))
    Permission.objects.get_or_create(label="file.download_source")
    Permission.objects.get_or_create(label="file.upload")


def set_file_types():
    for title in ["text", "numeric", "json", "xml", "integer", "float"]:
        MetadataType.objects.get_or_create(title=title)
    with open(
        "{}/mimes.json".format(pathlib.Path(__file__).parent.absolute()), "r"
    ) as f:
        data = json.load(f)
        for key in data:
            if "extensions" in data[key]:
                file_type, created = FileType.objects.get_or_create(mime=key, title=key)
                for extension in data[key]["extensions"]:
                    extension, created = FileExtension.objects.get_or_create(
                        label=extension
                    )
                    file_type.extensions.add(extension)
    for mime in IIIF_SUPPORT:
        try:
            file_type = FileType.objects.get(mime=mime)
            file_type.serve_with_iiif = True
            file_type.save()
        except FileType.DoesNotExist:
            pass


# deprecate ?
def set_basic_vocabularies_metas(project: Project):
    metadatasets = {}
    with open(
        "{}/vocabulary.csv".format(pathlib.Path(__file__).parent.absolute()), "r"
    ) as f:
        reader = csv.reader(f)
        next(reader)
        for row in reader:
            metadatasets[row[0]], created = MetadataSet.objects.get_or_create(
                title=row[4], project=project
            )
    metadatas = {}
    with open(
        "{}/property.csv".format(pathlib.Path(__file__).parent.absolute()), "r"
    ) as f:
        reader = csv.reader(f)
        next(reader)
        for row in reader:
            metadatas[row[0]], created = Metadata.objects.get_or_create(
                title=row[3],
                set_id=metadatasets[row[2]].pk,
                project=project,
            )
    with open(
        "{}/resource_class.csv".format(pathlib.Path(__file__).parent.absolute()),
        "r",
    ) as f:
        reader = csv.reader(f)
        next(reader)
        for row in reader:
            ResourceClass.objects.get_or_create(
                id=row[0],
                title=row[3],
                set_id=metadatasets[row[2]].pk,
                verbose_title=row[4],
                description=row[5],
            )


def load_fixtures(*args):
    set_base_permissions()
    set_file_types()


class Command(BaseCommand):
    def handle(self, *args, **options):
        load_fixtures()



# Generated by Django 5.0.1 on 2024-04-04 08:34

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0032_usertask"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="usertask",
            options={"ordering": ["-updated_at"]},
        ),
    ]



# Generated by Django 5.0.1 on 2024-04-18 12:35

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0035_remove_file_post_treated_project_use_exiftool"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="collection",
            options={
                "ordering": ("title",),
                "verbose_name": "collection",
                "verbose_name_plural": "collections",
            },
        ),
        migrations.AlterModelOptions(
            name="resource",
            options={
                "ordering": ("title",),
                "verbose_name": "ressource",
                "verbose_name_plural": "ressources",
            },
        ),
    ]



from pathlib import Path
import os
import tempfile
from shutil import which
from dotenv import load_dotenv
import sys
import secrets
import string

for tool in [
    "vips",
    "convert",
    "exiftool",
    "pdftoppm",
    "pdftotext",
    "tesseract",
]:
    if not Path(which(tool) or "/does/not/exist").exists():
        sys.exit(f"Couldn't find tool {tool}, please read installation instructions.")

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
BASE_DIR = Path(__file__).resolve().parent.parent

VAR_DIR = os.getenv("JAMA_VAR_DIR") or Path(Path.home(), ".jama")
VAR_DIR = Path(VAR_DIR).resolve()

os.makedirs(VAR_DIR, exist_ok=True)
env_file = Path(VAR_DIR, "env")
if not env_file.exists():
    secret = "".join(secrets.choice(string.ascii_letters) for _ in range(32))
    with open(env_file, "w") as f:
        f.write(f"JAMA_DEBUG=0\nJAMA_SECRET={secret}")
load_dotenv(env_file)

JAMA_DB_ADAPTER = os.getenv("JAMA_DB_ADAPTER") or "sqlite"
JAMA_URL_BASE_PATH = os.getenv("JAMA_URL_BASE_PATH") or ""
JAMA_USE_MODSHIB = True if os.getenv("JAMA_USE_MODSHIB", "1") == "1" else False
JAMA_USE_CACHALOT = True if os.getenv("JAMA_USE_CACHALOT", "0") == "1" else False
JAMA_USE_WEBP = True if os.getenv("JAMA_USE_WEBP", "1") == "1" else False
JAMA_USE_OAI = True if os.getenv("JAMA_USE_OAI", "0") == "1" else False
JAMA_USE_ARK = True if os.getenv("JAMA_USE_ARK", "0") == "1" else False
JAMA_SQLITE_DB_PATH = os.getenv("JAMA_SQLITE_DB_PATH") or Path(VAR_DIR, "db.sqlite3")
JAMA_IIIF_PROCESSING_DIR = os.getenv(
    "JAMA_IIIF_PROCESSING_DIR", Path(VAR_DIR, "processing")
)
PARTIAL_UPLOADS_DIR = os.getenv("JAMA_PARTIAL_UPLOADS_DIR") or str(
    Path(VAR_DIR, "partial_uploads").resolve()
)
os.makedirs(PARTIAL_UPLOADS_DIR, exist_ok=True)

MEDIA_FILES_DIR = os.getenv("JAMA_FILES_DIR") or str(
    Path(VAR_DIR, "media_source_files").resolve()
)
os.makedirs(MEDIA_FILES_DIR, exist_ok=True)

TMP_THUMBNAILS_DIR = os.getenv("JAMA_TMP_THUMBNAILS_DIR") or tempfile.gettempdir()

JAMA_SITE = os.getenv("JAMA_SITE") or "http://localhost:8000/"

JAMA_IIIF_UPSCALING_PREFIX = os.getenv("JAMA_IIIF_UPSCALING_PREFIX") or ""
IIIF_DIR = os.getenv("JAMA_IIIF_DIR") or str(Path(VAR_DIR, "iiif").resolve())
os.makedirs(IIIF_DIR, exist_ok=True)
IIIF_PATH_SEPARATOR = os.getenv("JAMA_IIIF_PATH_SEPARATOR") or os.path.sep


ARK_SERVER = os.getenv("JAMA_ARK_SERVER", "")
ARK_APP_ID = os.getenv("JAMA_ARK_APP_ID", "")
ARK_SECRET_KEY = os.getenv("JAMA_ARK_SECRET_KEY", "")
CACHE_FILE_LOCATION = os.getenv("CACHE_LOCATION", str(Path(VAR_DIR, "cache").resolve()))
CACHE_FILE_LOCATION = Path(CACHE_FILE_LOCATION).resolve()
os.makedirs(CACHE_FILE_LOCATION, exist_ok=True)

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = (
    os.getenv("JAMA_SECRET") or "7d*_8c!d$vv963qpr45_x)@f2t-x6fu2&yi+m+d6s!p!lt+_j+"
)

# SECURITY WARNING: don't run with debug  on in production!
DEBUG = True if os.getenv("JAMA_DEBUG") == "1" else False

JAMA_IIIF_ENDPOINT = os.getenv("JAMA_IIIF_ENDPOINT") or "http://localhost/iip/IIIF="
ALLOWED_HOSTS = ["*"]
CSRF_TRUSTED_ORIGINS = [JAMA_SITE[:-1]]

# Application definition

INSTALLED_APPS = [
    "corsheaders",
    "resources.apps.ResourcesConfig",
    "rpc.apps.RpcConfig",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "huey.contrib.djhuey",
    "authcli.apps.AuthcliConfig",
    "annotations.apps.AnnotationsConfig",
    "jama.apps.JamaConfig",
    "django_extensions",
]


if JAMA_USE_CACHALOT:
    INSTALLED_APPS.append("cachalot")

if JAMA_USE_MODSHIB:
    MODSHIB_SHOW_LOCAL_LOGIN = (
        True if os.getenv("MODSHIB_SHOW_LOCAL_LOGIN", "1") == "1" else False
    )
    MODSHIB_SHOW_SSO_LOGIN = (
        True if os.getenv("MODSHIB_SHOW_SSO_LOGIN", "0") == "1" else False
    )
    INSTALLED_APPS.append("modshib.apps.ModshibConfig")


AUTO_REGISTER_APPS = []
STATICFILES_DIRS = []
for app in os.getenv("JAMA_APPS", "").split(","):
    if app:
        if app == "bibnum":
            INSTALLED_APPS.append("django_vite")
        INSTALLED_APPS.append(f"{app}.apps.{app.capitalize()}Config")
        AUTO_REGISTER_APPS.append(app)


MIDDLEWARE = [
    "corsheaders.middleware.CorsMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

CORS_ALLOW_ALL_ORIGINS = True

CORS_ALLOW_HEADERS = [
    "accept",
    "accept-encoding",
    "authorization",
    "content-type",
    "dnt",
    "origin",
    "user-agent",
    "x-csrftoken",
    "x-requested-with",
    "X-Api-Key",
    "X-file-chunk",
    "X-file-hash",
    "X-file-name",
    "X-origin-dir",
    "X-Project",
]

ROOT_URLCONF = "jama.urls"

context_processors = [
    "django.template.context_processors.debug",
    "django.template.context_processors.request",
    "django.contrib.auth.context_processors.auth",
    "django.contrib.messages.context_processors.messages",
]

if JAMA_USE_MODSHIB:
    context_processors.append("modshib.context_processors.modshib_context")

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": context_processors,
        },
    },
]

WSGI_APPLICATION = "jama.wsgi.application"


if JAMA_DB_ADAPTER == "sqlite":
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": JAMA_SQLITE_DB_PATH,
            "OPTIONS": {
                "init_command": (
                    "PRAGMA foreign_keys=ON;"
                    "PRAGMA journal_mode = WAL;"
                    "PRAGMA synchronous = NORMAL;"
                    "PRAGMA busy_timeout = 5000;"
                    "PRAGMA temp_store = MEMORY;"
                    "PRAGMA mmap_size = 134217728;"
                    "PRAGMA journal_size_limit = 67108864;"
                    "PRAGMA cache_size = 2000;"
                ),
                "transaction_mode": "IMMEDIATE",
            },
        }
    }
else:
    DATABASES = {
        "default": {
            "ENGINE": "django.db.backends.postgresql",
            "HOST": os.getenv("JAMA_DB_HOST") or "localhost",
            "PORT": os.getenv("JAMA_DB_PORT") or "5432",
            "NAME": os.getenv("JAMA_DB_NAME") or "django",
            "USER": os.getenv("JAMA_DB_USER") or "django",
            "PASSWORD": os.getenv("JAMA_DB_PASSWORD") or "django",
        }
    }


AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


LANGUAGE_CODE = "fr-fr"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


STATIC_ROOT = os.getenv("JAMA_STATIC_ROOT") or Path(VAR_DIR, "static")
STATIC_URL = JAMA_URL_BASE_PATH + "/static/"

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "console": {"format": "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"},
        "file": {"format": "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"},
    },
    "handlers": {
        "console": {
            "level": "INFO",
            "class": "logging.StreamHandler",
            "formatter": "console",
        },
    },
    "loggers": {
        "": {"level": "INFO", "handlers": ["console"]},
    },
}


HUEY = {
    "huey_class": "huey.SqliteHuey",  # Huey implementation to use.
    "filename": Path(VAR_DIR, "huey.db").resolve(),
    "results": True,  # Store return values of tasks.
    "store_none": False,  # If a task returns None, do not save to results.
    "immediate": False,  # If DEBUG=True, run synchronously.
    "utc": True,  # Use UTC for all times internally.
    "consumer": {
        "workers": 1,
        "worker_type": "thread",
        "initial_delay": 0.1,  # Smallest polling interval, same as -d.
        "backoff": 1.15,  # Exponential backoff using this rate, -b.
        "max_delay": 10.0,  # Max possible polling interval, -m.
        "scheduler_interval": 1,  # Check schedule every second, -s.
        "periodic": True,  # Enable crontab feature.
        "check_worker_health": True,  # Enable worker health checks.
        "health_check_interval": 1,  # Check worker health every second.
    },
}


CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
        "LOCATION": CACHE_FILE_LOCATION,
    }
}

LOGIN_REDIRECT_URL = "/ui/"

if "bibnum" in AUTO_REGISTER_APPS:
    DJANGO_VITE_ASSETS_PATH = Path(BASE_DIR, "static", "bibnum")
    DJANGO_VITE_DEV_MODE = DEBUG
    DJANGO_VITE_MANIFEST_PATH = Path(BASE_DIR, "static", "bibnum", "manifest.json")

if "ui" in AUTO_REGISTER_APPS:
    STATICFILES_DIRS.append(Path(BASE_DIR, "ui", "front", "dist"))



# Generated by Django 5.0.1 on 2024-02-08 10:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("resources", "0029_project_ark_redirect"),
    ]

    operations = [
        migrations.AddField(
            model_name="file",
            name="denormalized_image_height",
            field=models.IntegerField(
                blank=True, null=True, verbose_name="hauteur de l'image"
            ),
        ),
        migrations.AddField(
            model_name="file",
            name="denormalized_image_width",
            field=models.IntegerField(
                blank=True, null=True, verbose_name="largeur de l'image"
            ),
        ),
    ]



from django.http import HttpRequest, HttpResponse, JsonResponse
from django.views.decorators.http import require_http_methods
from django.views.decorators.csrf import csrf_exempt
from django.contrib.auth.models import User
from django.contrib.auth import authenticate
from django.shortcuts import render
from django import forms
import json
import rpc.methods as rpc_methods
from rpc.methods import ServiceException, _check_project_permission
from inspect import getmembers, isfunction
from resources.helpers import (
    handle_uploaded_file,
    UnknownFileType,
    ResourceError,
    ConcurrencyError,
)
from resources import models
from typing import Union
from jama import settings
import re
import os
from glob import glob
import hashlib
from django.core.files.base import File
from django.db import transaction
from pathlib import Path
from shutil import rmtree
import base64
from .fileresponse import RangedFileResponse
from django.views.decorators.gzip import gzip_page
import logging
from functools import lru_cache
from resources.tasks import (
    exif_task,
    iiif_task,
    ocr_task,
    update_resources_from_xlsx_rows,
)
from openpyxl import Workbook, load_workbook
import tempfile


importlib = __import__("importlib")

logger = logging.getLogger(__name__)


def _silent_rmdir(dir_path: str):
    rmtree(dir_path, ignore_errors=True)


def _silent_remove(file_path: str):
    try:
        os.remove(file_path)
    except FileNotFoundError:
        pass


def _file_hash256(file_path: str) -> str:
    hsh = hashlib.sha256()
    with open(file_path, "rb") as f:
        for chnk in iter(lambda: f.read(8192), b""):
            hsh.update(chnk)
    return hsh.hexdigest()


def _collection_from_origin_dir(
    origin_dir: str, project: models.Project
) -> Union[models.Collection, None]:
    # root dir always first
    previous_dir = project.root_collection
    for dir_name in origin_dir.split("/"):
        dir_name = dir_name.strip()
        if not dir_name:
            continue
        dir_name = dir_name.strip()
        if dir_name:
            previous_dir, created = models.Collection.objects.get_or_create(
                project=project, title=dir_name, parent=previous_dir
            )
            # this was soft-deleted. Un-delete !
            if not created and previous_dir.deleted_at:
                previous_dir.deleted_at = None
                previous_dir.save()
    return previous_dir


def _join_parts(
    parts_dir: str, destination: str, sha256sum: str, total_parts: int
) -> bool:
    with open(destination, "wb") as f:
        for path in sorted(glob("{}/{}-*.part".format(parts_dir, total_parts))):
            with open(path, "rb") as s:
                f.write(s.read())

    for path in glob("{}/{}-*.part".format(parts_dir, total_parts)):
        _silent_remove(path)
    if _file_hash256(destination) != sha256sum:
        return False
    return True


def _get_user_from_request(request: HttpRequest) -> Union[User, None]:
    """
    Tries to find a user given the request. In this order:

    - try to find session user
    - try to find user associated with X-api-key header
    - try to find user corresponding to Authorization header
    """
    if request.user.is_authenticated:
        logger.debug(f"user {request.user.username} is authenticated")
        return request.user
    if "X-Api-Key" in request.headers:
        try:
            return User.objects.get(
                apikey__key=request.headers["X-Api-Key"], apikey__active=True
            )
        except User.DoesNotExist:
            logger.warning("X-Api-Key was given but user was not found")
            return None
    if "Authorization" in request.headers:
        try:
            auth_u, auth_p = (
                base64.decodebytes(request.headers["Authorization"][6:].encode("utf-8"))
                .decode("utf-8")
                .split(":")
            )
            user = authenticate(username=auth_u, password=auth_p)
            if not user:
                logger.warning(
                    "Authorization was given in request headers but user was not authenticated"
                )
            return user
        except KeyError:
            logger.warning("Authorization was malformed")
            return None


def _get_project_from_request(request: HttpRequest) -> Union[models.Project, None]:
    if "X-Project" in request.headers:
        return models.Project.objects.filter(pk=request.headers["X-Project"]).first()
    return None


@lru_cache(None, typed=True)
def _get_rpc_methods() -> dict:
    methods = {}
    for fn_name, _ in getmembers(rpc_methods, isfunction):
        if fn_name[0:1] == "_":
            continue
        methods[fn_name] = getattr(rpc_methods, fn_name)
    for app_name in settings.AUTO_REGISTER_APPS:
        try:
            rpc_module = importlib.import_module("{}.rpc.methods".format(app_name))
            for fn_name, _ in getmembers(rpc_module, isfunction):
                if fn_name[0:1] == "_":
                    continue
                methods[fn_name] = getattr(rpc_module, fn_name)
        except ModuleNotFoundError:
            continue
    return methods


@gzip_page
@transaction.atomic
@csrf_exempt
def rpc(request: HttpRequest) -> HttpResponse:
    req_id = None
    if request.method == "GET":
        return render(request, "docs.html", {"methods": _get_rpc_methods()})
    if request.method != "POST":
        return HttpResponse("Method Not Allowed", status=405)
    try:
        user = _get_user_from_request(request)
        if not user:
            return HttpResponse("Forbidden", status=403)
        methods = _get_rpc_methods()
        json_data = json.loads(request.body)
        method = json_data["method"]
        params = json_data["params"] if "params" in json_data else None
        if not params:
            params = []
        params.insert(0, user)  # always add user to method call
        req_id = json_data["id"]
        result = methods[method](*params)
        return JsonResponse({"result": result, "error": None, "id": req_id})
    except json.JSONDecodeError:
        logger.warning("could not decode json request")
        return HttpResponse("Bad Request", status=400)
    except KeyError:
        logger.warning("client data dict key error")
        return HttpResponse("Bad Request", status=400)
    except TypeError:
        logger.warning("client data type error")
        return HttpResponse("Bad Request", status=400)
    except ValueError:
        logger.warning("client data warning error")
        return HttpResponse("Bad Request", status=400)
    except ServiceException as e:
        return JsonResponse(
            {
                "result": None,
                "error": getattr(e, "message", repr(e)),
                "id": req_id,
            }
        )


@csrf_exempt
@require_http_methods(["POST"])
def upload_partial(request: HttpRequest) -> HttpResponse:
    user = _get_user_from_request(request)
    if not user:
        return HttpResponse("Forbidden", status=403)
    project = _get_project_from_request(request)
    if not project:
        return HttpResponse("Bad Request", status=400)
    try:
        _check_project_permission(user, project, "file.upload")
    except ServiceException:
        return HttpResponse("Forbidden", status=403)
    # optional header, used to create collections
    origin_dir = request.headers.get("X-origin-dir", None)

    try:
        file_hash = request.headers["X-file-hash"]
        if re.match("[A-Fa-f0-9]{64}", file_hash) is None:
            raise ValueError
        file_name = base64.b64decode(request.headers["X-file-name"]).decode("utf-8")
        _, f_extension = os.path.splitext(file_name)
        if not f_extension:
            return HttpResponse("Type de fichier inconnu", status=400)
        chunk_number, total_chunks = request.headers["X-file-chunk"].split("/")
        file_instance = models.File.objects.get(hash=file_hash, project=project)
        # Rise from your grave !
        if file_instance.deleted_at:
            file_instance.deleted_at = None
            file_instance.title = file_name
            file_instance.original_name = file_name
            file_instance.save()
            logger.info(
                "User {}({}) respawned file {}".format(
                    user.username, user.pk, file_instance.pk
                )
            )
        if origin_dir:
            collection = _collection_from_origin_dir(origin_dir, project)
            collection.resources.add(file_instance)
            logger.info(
                "User {}({}) added file({}) to collection({})".format(
                    user.username, user.pk, file_instance.pk, collection.pk
                )
            )
        return HttpResponse(file_instance.id, status=200)
    except (KeyError, ValueError):
        return HttpResponse(status=400)
    except models.File.DoesNotExist:
        pass

    partials_dir = "{}/{}-{}".format(settings.PARTIAL_UPLOADS_DIR, user.pk, file_hash)
    lock_file_path = "{}/.lock".format(partials_dir)
    if os.path.isfile(lock_file_path):
        return HttpResponse(status=202)
    os.makedirs(partials_dir, exist_ok=True)

    # We don't want to store an incomplete chunk.
    # (prevent client disconnect)
    try:
        if len(request.body) == int(request.headers["Content-Length"]):
            partial_destination = "{}/{}-{}.part".format(
                partials_dir, total_chunks, chunk_number.zfill(len(total_chunks))
            )
            with open(partial_destination, "wb") as f:
                f.write(request.body)
        else:
            return HttpResponse("Content-Length does not match body size", status=400)
    except KeyError:  # Content-length header is mandatory
        return HttpResponse("Length Required", status=411)

    # All chunks are complete, time to assemble
    f_name = "{}/complete{}".format(partials_dir, f_extension)
    if len(glob("{}/{}-*.part".format(partials_dir, total_chunks))) == int(
        total_chunks
    ) or os.path.exists(f_name):
        Path(lock_file_path).touch()  # lock dir

        if not os.path.exists(f_name):
            checksum_ok = _join_parts(partials_dir, f_name, file_hash, total_chunks)
        elif file_hash == _file_hash256(f_name):
            checksum_ok = True
        else:
            checksum_ok = False
        try:
            with open(f_name, "rb") as f:
                try:
                    file_id = None
                    with transaction.atomic():
                        file_id = handle_uploaded_file(
                            File(f),
                            project,
                            force_file_name=file_name,
                        )
                        f.close()
                        _silent_rmdir(partials_dir)
                        logger.info(
                            "User {}({}) handled file {}".format(
                                user.username, user.pk, file_id
                            )
                        )
                        if origin_dir:
                            collection = _collection_from_origin_dir(
                                origin_dir, project
                            )
                            collection.resources.add(
                                models.File.objects.get(id=file_id)
                            )
                            logger.info(
                                "User {}({}) added file({}) to collection({})".format(
                                    user.username, user.pk, file_id, collection.pk
                                )
                            )
                    iiif_task(file_id)
                    exif_task(file_id)
                    ocr_task(file_id)
                    if not checksum_ok:
                        return HttpResponse(file_id, status=210)  # content different
                    else:
                        return HttpResponse(file_id, status=200)
                except UnknownFileType:
                    _silent_rmdir(partials_dir)
                    return HttpResponse("Type de fichier inconnu", status=400)
                except ResourceError:
                    _silent_rmdir(partials_dir)
                    return HttpResponse(
                        "Impossible d'enregistrer la ressource", status=400
                    )
                except ConcurrencyError:
                    # Concurrency problem with clients sending multiple chunks in parallel.
                    # Nothing to do here, work has already been done.
                    return HttpResponse("Too Many Requests", status=429)
        # another race condition
        except FileNotFoundError:
            return HttpResponse("Too Many Requests", status=429)
    return HttpResponse(status=202)


def force_download(
    request: HttpRequest, file_id: int
) -> Union[HttpResponse, RangedFileResponse]:
    file = models.File.objects.filter(id=file_id).first()
    if not file:
        return HttpResponse("Not Found", 404)
    try:
        if not settings.DEBUG:  # bypass auth in dev
            user = _get_user_from_request(request)
            if not user:
                return HttpResponse("Forbidden", status=403)
            _check_project_permission(user, file.project, "file.download_source")
    except ServiceException:
        return HttpResponse("Forbidden", status=403)
    return RangedFileResponse(
        request,
        open(file.local_path(), "rb"),
        filename=file.new_filename,
        # content_type=file.file_type.mime, # Firefox is bad
    )


def force_download_public(
    request: HttpRequest, file_hash: str
) -> Union[HttpResponse, RangedFileResponse]:
    file = models.File.objects.filter(hash=file_hash).first()
    if not file:
        return HttpResponse("Not Found", 404)
    return RangedFileResponse(
        request,
        open(file.local_path(), "rb"),
        filename=file.new_filename,
        # content_type=file.file_type.mime, # Firefox is bad
    )


@csrf_exempt
@require_http_methods(["POST"])
def upload_metas_xls(request: HttpRequest) -> HttpResponse:
    data_rows = []
    user = _get_user_from_request(request)
    if not user:
        return HttpResponse("Forbidden", status=403)

    class MetasXSLXUploadForm(forms.Form):
        file = forms.FileField()

    form = MetasXSLXUploadForm(request.POST, request.FILES)
    if form.is_valid():
        with tempfile.NamedTemporaryFile(suffix=".xlsx") as tmp:
            with open(tmp.name, "wb+") as destination:
                for chunk in request.FILES["file"].chunks():
                    destination.write(chunk)
            wb = load_workbook(filename=tmp)
            ws = wb.active
            keys = []
            for row in ws.iter_rows(values_only=True):
                if not keys:
                    keys = list(row)
                else:
                    data = dict(zip(keys, list(row)))
                    data_rows.append(data)
            user_task = models.UserTask.objects.create(
                owner=user, description="Traitement de fichier XLS de métadonnées"
            )
            update_resources_from_xlsx_rows(user.pk, data_rows, user_task.pk)
            return HttpResponse("OK", status=200)
    return HttpResponse("Bad Request", status=400)


def download_metas_xls(request: HttpRequest, collection_id):
    collection_instance = models.Collection.objects.filter(pk=collection_id).first()
    if not collection_instance:
        return HttpResponse("Not Found", status=404)
    if not settings.DEBUG:  # bypass auth in dev
        user = _get_user_from_request(request)
        if not user:
            return HttpResponse("Forbidden", status=403)
        _check_project_permission(user, collection_instance.project, "collection.read")

    wb = Workbook()
    ws = wb.active
    row_number = 2  # /!\ Index starts at 1. Leave the first row for headers.
    for r in collection_instance.yield_resource_data_for_export():
        col_number = 1
        for value in r.values():
            cell = ws.cell(row=row_number, column=col_number)
            cell.value = value
            col_number = col_number + 1
        row_number = row_number + 1
    col_number = 1
    for key in r.keys():
        cell = ws.cell(row=1, column=col_number)
        cell.value = key
        col_number = col_number + 1

    with tempfile.NamedTemporaryFile(suffix=".xlsx") as tmp:
        wb.save(tmp.name)
        return RangedFileResponse(
            request,
            open(tmp.name, "rb"),
            filename=f"collection_{collection_id}.xlsx",
            content_type="application/vnd.ms-excel",
        )



from django.core.management.base import BaseCommand
from resources.models import Project, ProjectAccess, Role
from django.contrib.auth.models import User


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("user", type=str)
        parser.add_argument("project", type=str)
        parser.add_argument("role", type=str)

    def handle(self, *args, **options):
        user = User.objects.get(username=options["user"])
        project = Project.objects.get(label=options["project"])
        role = Role.objects.get(project=project, label=options["role"])
        ProjectAccess.objects.get_or_create(role=role, project=project, user=user)
